%macro get_eigens(
dsn=, 
	/*** Input data set name. This must contain columns named
	locus_no, corr_0_05, corr_05_1, corr_0_1, and max_abs_zs ***/
out_eigen_vecs=eigen_vecs
	/*** Output data set containing eigen_vecs.  Note that if this data set
	isn't in library "WORK", the library should be affixed to the data set name. ***/
);
/*** Note that, in this MACRO, a data set called "eigen_vals" is created and destroyed. ***/  
/*** 
*** Calculate eigenvalues, calculate eigenvectors, count number of distinct SNPs ***
***/
/***********************************************************************/

/*** Calculate eigenvalues ***/

/***********************************************************************/
data eigen_vals(keep=locus_no eigen_val1 eigen_val2 eigen_val3 ); 
set &dsn.(keep=locus_no corr_0_05 corr_05_1 corr_0_1);
/*** Referencing http://mathworld.wolfram.com/CubicFormula.html ***/
/*** Define constants ***/
pi=constant('pi'); round_unit=1e-8; 
p=-(corr_0_05**2+corr_05_1**2+corr_0_1**2); q=-2*corr_0_05*corr_05_1*corr_0_1;
big_Q=p/3; big_R=q/2; theta=arcos(big_R/sqrt(-1*big_Q**3));
eigen_val1=round(1-2*sqrt(-1*big_Q)*cos(theta/3), round_unit);
pi=constant('pi');
eigen_val2=round(1-2*sqrt(-1*big_Q)*cos((theta+2*pi)/3), round_unit);
eigen_val3=round(1-2*sqrt(-1*big_Q)*cos((theta+4*pi)/3), round_unit);
run;

/***********************************************************************/

/*** Calculate eigenvectors ***/

/***********************************************************************/
data &out_eigen_vecs.(drop=tmp one i j k fac b1 b2 b3 summ 
	tmp_mat1 tmp_mat2 tmp_mat3 tmp_mat4 tmp_mat5 tmp_mat6 tmp_mat7 tmp_mat8 tmp_mat9 ); 
merge eigen_vals(in=A) 
	&dsn.(keep=locus_no corr_0_05 corr_05_1 corr_0_1 max_abs_zs in=B)
	end=last;
if a and b;
one=1;
array cor_mat(1:3, 1:3) one corr_0_05 corr_0_1 corr_0_05 one corr_05_1 corr_0_1 corr_05_1 one ;
array eig_vals(1:3) eigen_val1 eigen_val2 eigen_val3 ;
array tmp_mat(1:3, 1:3) ;
* array track_changes(1:3) one two three;
/* Determine how many vectors we'll find and sort from largest to smallest */
/* Note that the corr mats are at least semi-pos definite so smallest possible
eig val is zero */
/* Put eigen vals in descending order. */
do i =1 to 2;
	if eig_vals[i+1]>eig_vals[i] then do;
		tmp=eig_vals[i]; eig_vals[i]=eig_vals[i+1]; eig_vals[i+1]=tmp;
	end;
end;
/* if (middle, smallest, biggest) or (smallest, middle, biggest) a 3rd step is needed */
if eig_vals[2]>eig_vals[1] then do;
		tmp=eig_vals[1]; eig_vals[1]=eig_vals[2]; eig_vals[2]=tmp;
end;
/* Find the eigen vecs */
do i =1 to 3;
if eig_vals[i]>0 then do;
	/* Load the values from cor_mat into tmp_mat */ 
	do j =1 to 3;
		do k =1 to 3;
			tmp_mat[j,k]=cor_mat[j,k];
		end;
	end;
	tmp_mat[1,1]=1-eig_vals[i]; tmp_mat[2,2]=tmp_mat[1,1]; tmp_mat[3,3]=tmp_mat[1,1];
	/* Do the first row */
	/* See if we need any row exchanges to avoid using a bad pivot.  */
	do j =1 to 2;
		if abs(tmp_mat[j+1,1])>abs(tmp_mat[j,1]) then do;
			do k =1 to 3;
				tmp=tmp_mat[j,k]; tmp_mat[j,k]=tmp_mat[j+1,k]; tmp_mat[j+1,k]=tmp; 
			end;
		end;
	end;
	if abs(tmp_mat[2,1])>abs(tmp_mat[1,1]) then do;
		do k =1 to 3;
			tmp=tmp_mat[1,k]; tmp_mat[1,k]=tmp_mat[2,k]; tmp_mat[2,k]=tmp;
		end;
	end;
	/* Do the subtraction. We only need to subtract multiple of 1st row from 2nd row */
	/* Get the factor for the subtraction */
	fac=tmp_mat[2,1]/tmp_mat[1,1]; 
	tmp_mat[2,1]=0; 
	do j = 2 to 3;
		tmp_mat[2,j]=tmp_mat[2,j]-fac*tmp_mat[1,j];
	end;
	/* For the moment choose b3=1 and solve the other 2 */
	b3=1;
	b2=-1*tmp_mat[2,3]/tmp_mat[2,2];
	b1=(tmp_mat[2,3]*tmp_mat[1,2]/tmp_mat[2,2]-tmp_mat[1,3])/tmp_mat[1,1];
	/* Rescale bs */
	fac=sqrt(b1**2+b2**2+b3**2);
	b1=b1/fac; b2=b2/fac; b3=b3/fac;
	/* Load the b's on the apropriate vars */ 
	if i=1 then do; b1_1=b1; b2_1=b2; b3_1=b3; end;
	if i=2 or i=3 then do; 
		/* Make sure that eigen vecs are orthogonal */
		summ=round(b1_1*b1+b2_1*b2+b3_1*b3, 1e-6);
		if summ^=0 then do; 
			b1=-1*b1; b2=-1*b2; b3=-1*b3; 
		end;
		if i=2 then do;
			b1_2=b1; b2_2=b2; b3_2=b3; 
		end;
		if i=3 then do;
			b1_3=b1; b2_3=b2; b3_3=b3;
		end;
	end; /* End: 'if i=2 or i=3 then do;' */
end; /* End of 'if eig_vals[i]>0 then do;' */
end; /* End of 'Find the eigen vecs' ... do i =1 to 3; */

run;

/*** Clean up of data sets. ***/
proc datasets; delete EIGEN_ValS; run; 

%mend;
